#%%
from collocation import Collocation

# Prepare corpus data
# https://yongfu.name/collocation/sampled_PTTposts.txt
corpus = []
with open("sampled_PTTposts.txt", encoding="utf-8") as f:
    for sent in f.read().split("\n"):
        if sent.strip() == "": continue
        sentence = []
        for tk in sent.split("\u3000"):
            if tk == "": continue
            sentence.append(tk)
        corpus.append(sentence)

corpus[:7]

# Initialize
c = Collocation(corpus, left_window=3, right_window=3)
# Query
c.get_topn_collocates("[臺台]灣", cutoff=3, n=3, by="MI", chinese_only=True)

# Acess documentation of parameters
help(c.get_topn_collocates)


# %%
from math import log2
from collocation.association import FisherAttract, Dice


def logDice(O11, O12, O21, O22, E11, E12, E21, E22):
    D = Dice(O11, O12, O21, O22, E11, E12, E21, E22)
    return 14 + log2(D)

c.association_measures = [FisherAttract, logDice]
c.get_topn_collocates("[臺台]灣", cutoff=3, n=3, by="logDice", chinese_only=True)
# %%



import re
from .association import MI, Xsq, Gsq, Dice, DeltaP21, DeltaP12, additive_smooth

PAT_CH_CHR = re.compile("[〇一-\u9fff㐀-\u4dbf豈-\ufaff]")


class Collocation:
    association_measures = [
        MI, Xsq, Gsq, Dice, DeltaP21, DeltaP12
    ]

    def __init__(self, corpus, left_window, right_window):
        self.left_window = left_window
        self.right_window = right_window
        self.cooccur_freq = {}
        self.R1_all = {}  # new
        self.C1_all = {}  # new
        self.count_freq(corpus)
        self.N = sum(self.cooccur_freq.values())  # new
    

    def get_topn_collocates(self, node, cutoff, n=10, alpha=0, 
                            chinese_only=True, by="MI"):
        """Retrieve top n collocates of a node

        Parameters
        ----------
        node : str
            Regex pattern describing the node word
        cutoff : int
            The minimum number of coocurrences between the node and a collocate
            needed for the pair to be assign association scores
        n : int, optional
            The number of collocates to return, by default 10
        alpha : int, optional
            Additive smoothing parameter, by default 0. This is usually set
            between 0 and 1
        by : str, optional
            The association measure used for sorting, by default "MI"
            Possible values are the names of the functions in
            :meth:`Collocation.association_measures`
        chinese_only : bool, optional
            Whether to return only collocates that have at least one Chinese
            character, by default True

        Returns
        -------
        list
            A list of tuples (node, collocate, associations)
        """
        node = re.compile(node)
        candidates = []
        for node_, collo_ in self.cooccur_freq:
            if chinese_only:
                if not any(PAT_CH_CHR.match(c) for c in collo_): continue
            if node.match(node_):
                asso = self.association(node=node_, collocate=collo_, 
                                        cutoff=cutoff, alpha=alpha)
                if asso is not None:
                    candidates.append( (node_, collo_, asso) )
        if isinstance(n, int):
            return sorted(candidates, key=lambda x: x[-1][by], reverse=True)[:n]
        else:
            return sorted(candidates, key=lambda x: x[-1][by], reverse=True)


    def association(self, node, collocate, cutoff, alpha):
        # Retrieve frequencies
        O11 = self.cooccur_freq.get((node, collocate))
        R1 = self.R1_all.get(node)
        C1 = self.C1_all.get(collocate)
        if O11 is None or R1 is None or C1 is None: return None
        if O11 < cutoff: return None
        R2 = self.N - R1
        O12 = R1 - O11
        O21 = C1 - O11
        O22 = R2 - O21
        O11, O12, O21, O22, E11, E12, E21, E22 = additive_smooth(O11, O12, O21, O22, alpha=alpha)

        stats = { 
                func.__name__: func(O11, O12, O21, O22, E11, E12, E21, E22)\
                    for func in self.association_measures
        }
        stats['RawCount'] = O11

        return stats


    def count_freq(self, sentences):
        for sent in sentences: 
            self.count_freq_sent(sent)


    def count_freq_sent(self, sentence):
        sent_len = len(sentence)
        
        # Sliding window (跑過句子中的每個詞彙)
        for i, node in enumerate(sentence):
            
            # Set window to scan through
            win_left = max(i - self.left_window, 0)
            win_right = min(i + self.right_window + 1, sent_len)
            collocates = sentence[win_left:i] + sentence[i+1:win_right]
        
            for collocate in collocates:
                # Count cooccurance frequency
                k = (node, collocate)
                if k not in self.cooccur_freq:
                    self.cooccur_freq[k] = 0
                self.cooccur_freq[k] += 1

                # Count marginal frequencies
                # Count node marginal frequency
                if node not in self.R1_all: 
                    self.R1_all[node] = 0
                self.R1_all[node] += 1

                # Count collocate marginal frequency
                if collocate not in self.C1_all:
                    self.C1_all[collocate] = 0
                self.C1_all[collocate] += 1



# Efficient Fisher exact test implementation by painyeph/FishersExactTest
# https://github.com/painyeph/FishersExactTest/blob/master/fisher.py
#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: eph

from math import log, exp

try:

    from math import lgamma  # Python 2.7+

except ImportError:

    _cache = {0: 0., 1: 0.}
    def lgamma(z):
        if z in _cache: return _cache[z]
        x = 0.1659470187408462e-06 / (z + 7)
        x += 0.9934937113930748e-05 / (z + 6)
        x -= 0.1385710331296526 / (z + 5)
        x += 12.50734324009056 / (z + 4)
        x -= 176.6150291498386 / (z + 3)
        x += 771.3234287757674 / (z + 2)
        x -= 1259.139216722289 / (z + 1)
        x += 676.5203681218835 / z
        x += 0.9999999999995183
        _cache[z] = x = log(x) - 5.58106146679532777 - z + (z-0.5)*log(z+6.5)
        return x


def _maxn():
    l = 1; n = 2; h = float('inf')
    while l < n:
        if abs(lgamma(n+1) - lgamma(n) - log(n)) >= 1: h = n
        else: l = n
        n = (l + min(h, l * 3)) // 2
    return n

LN10 = log(10)
NINF = float('-inf')
MAXN = _maxn()


# ======================== Full Test ========================

def test1(a, b, c, d):
    result = mlnTest2(a, a+b, a+c, a+b+c+d)
    return exp(-result[0]), exp(-result[1]), exp(-result[2])

def test2(a, ab, ac, abcd):
    result = mlnTest2(a, ab, ac, abcd)
    return exp(-result[0]), exp(-result[1]), exp(-result[2])

def mlnTest1(a, b, c, d):
    return mlnTest2(a, a+b, a+c, a+b+c+d)

def mlnTest2(a, ab, ac, abcd):
    if 0 > a or a > ab or a > ac or ab+ac > abcd+a: raise ValueError('invalid contingency table')
    if abcd > MAXN: raise OverflowError('the grand total of contingency table is too large')
    a_min = max(0, ab+ac-abcd)
    a_max = min(ab, ac)
    if a_min == a_max: return 0., 0., 0.
    p0 = lgamma(ab+1) + lgamma(ac+1) + lgamma(abcd-ac+1) + lgamma(abcd-ab+1) - lgamma(abcd+1)
    pa = lgamma(a+1) + lgamma(ab-a+1) + lgamma(ac-a+1) + lgamma(abcd-ab-ac+a+1)
    sl = sr = 0.
    if ab * ac < a * abcd:
        for i in range(min(a-1, int(round(ab * ac / abcd))), a_min-1, -1):
            pi = lgamma(i+1) + lgamma(ab-i+1) + lgamma(ac-i+1) + lgamma(abcd-ab-ac+i+1)
            if pi < pa: continue
            sl_new = sl + exp(pa - pi)
            if sl_new == sl: break
            sl = sl_new
        for i in range(a+1, a_max+1):
            pi = lgamma(i+1) + lgamma(ab-i+1) + lgamma(ac-i+1) + lgamma(abcd-ab-ac+i+1)
            sr_new = sr + exp(pa - pi)
            if sr_new == sr: break
            sr = sr_new
        return -log(1.-max(0,exp(p0-pa)*sr)), max(0,pa-p0-log(1.+sr)), max(0,pa-p0-log(sl+1.+sr))
    else:
        for i in range(a-1, a_min-1, -1):
            pi = lgamma(i+1) + lgamma(ab-i+1) + lgamma(ac-i+1) + lgamma(abcd-ab-ac+i+1)
            sl_new = sl + exp(pa - pi)
            if sl_new == sl: break
            sl = sl_new
        for i in range(max(a+1, int(round(ab * ac / abcd))), a_max+1):
            pi = lgamma(i+1) + lgamma(ab-i+1) + lgamma(ac-i+1) + lgamma(abcd-ab-ac+i+1)
            if pi < pa: continue
            sr_new = sr + exp(pa - pi)
            if sr_new == sr: break
            sr = sr_new
        return max(0,pa-p0-log(sl+1.)), -log(1.-max(0,exp(p0-pa)*sl)), max(0,pa-p0-log(sl+1.+sr))

def mlog10Test1(a, b, c, d):
    result = mlnTest2(a, a+b, a+c, a+b+c+d)
    return result[0]/LN10, result[1]/LN10, result[2]/LN10

def mlog10Test2(a, ab, ac, abcd):
    result = mlnTest2(a, ab, ac, abcd)
    return result[0]/LN10, result[1]/LN10, result[2]/LN10


# ======================== Left Tail Only ========================

def test1l(a, b, c, d):
    return exp(-mlnTest2l(a, a+b, a+c, a+b+c+d))

def test2l(a, ab, ac, abcd):
    return exp(-mlnTest2l(a, ab, ac, abcd))

def mlnTest1l(a, b, c, d):
    return mlnTest2l(a, a+b, a+c, a+b+c+d)

def mlnTest2l(a, ab, ac, abcd):
    if 0 > a or a > ab or a > ac or ab+ac > abcd+a: raise ValueError('invalid contingency table')
    if abcd > MAXN: raise OverflowError('the grand total of contingency table is too large')
    a_min = max(0, ab+ac-abcd)
    a_max = min(ab, ac)
    if a_min == a_max: return 0.
    p0 = lgamma(ab+1) + lgamma(ac+1) + lgamma(abcd-ac+1) + lgamma(abcd-ab+1) - lgamma(abcd+1)
    pa = lgamma(a+1) + lgamma(ab-a+1) + lgamma(ac-a+1) + lgamma(abcd-ab-ac+a+1)
    if ab * ac < a * abcd:
        sr = 0.
        for i in range(a+1, a_max+1):
            sr_new = sr + exp(pa - lgamma(i+1) - lgamma(ab-i+1) - lgamma(ac-i+1) - lgamma(abcd-ab-ac+i+1))
            if sr_new == sr: break
            sr = sr_new
        return -log(1. - max(0, exp(p0 - pa) * sr))
    else:
        sl = 1.
        for i in range(a-1, a_min-1, -1):
            sl_new = sl + exp(pa - lgamma(i+1) - lgamma(ab-i+1) - lgamma(ac-i+1) - lgamma(abcd-ab-ac+i+1))
            if sl_new == sl: break
            sl = sl_new
        return max(0, pa - p0 - log(sl))

def mlog10Test1l(a, b, c, d):
    return mlnTest2l(a, a+b, a+c, a+b+c+d)/LN10

def mlog10Test2l(a, ab, ac, abcd):
    return mlnTest2l(a, ab, ac, abcd)/LN10


# ======================== Right Tail Only ========================

def test1r(a, b, c, d):
    return exp(-mlnTest2r(a, a+b, a+c, a+b+c+d))

def test2r(a, ab, ac, abcd):
    return exp(-mlnTest2r(a, ab, ac, abcd))

def mlnTest1r(a, b, c, d):
    return mlnTest2r(a, a+b, a+c, a+b+c+d)

def mlnTest2r(a, ab, ac, abcd):
    if 0 > a or a > ab or a > ac or ab+ac > abcd+a: raise ValueError('invalid contingency table')
    if abcd > MAXN: raise OverflowError('the grand total of contingency table is too large')
    a_min = max(0, ab+ac-abcd)
    a_max = min(ab, ac)
    if a_min == a_max: return 0.
    p0 = lgamma(ab+1) + lgamma(ac+1) + lgamma(abcd-ac+1) + lgamma(abcd-ab+1) - lgamma(abcd+1)
    pa = lgamma(a+1) + lgamma(ab-a+1) + lgamma(ac-a+1) + lgamma(abcd-ab-ac+a+1)
    if ab * ac > a * abcd:
        sl = 0.
        for i in range(a-1, a_min-1, -1):
            sl_new = sl + exp(pa - lgamma(i+1) - lgamma(ab-i+1) - lgamma(ac-i+1) - lgamma(abcd-ab-ac+i+1))
            if sl_new == sl: break
            sl = sl_new
        return -log(1. - max(0, exp(p0 - pa) * sl))
    else:
        sr = 1.
        for i in range(a+1, a_max+1):
            sr_new = sr + exp(pa - lgamma(i+1) - lgamma(ab-i+1) - lgamma(ac-i+1) - lgamma(abcd-ab-ac+i+1))
            if sr_new == sr: break
            sr = sr_new
        return max(0, pa - p0 - log(sr))

def mlog10Test1r(a, b, c, d):
    return mlnTest2r(a, a+b, a+c, a+b+c+d)/LN10

def mlog10Test2r(a, ab, ac, abcd):
    return mlnTest2r(a, ab, ac, abcd)/LN10


# ======================== Two Tails Only ========================

def test1t(a, b, c, d):
    return exp(-mlnTest2t(a, a+b, a+c, a+b+c+d))

def test2t(a, ab, ac, abcd):
    return exp(-mlnTest2t(a, ab, ac, abcd))

def mlnTest1t(a, b, c, d):
    return mlnTest2t(a, a+b, a+c, a+b+c+d)

def mlnTest2t(a, ab, ac, abcd):
    if 0 > a or a > ab or a > ac or ab+ac > abcd+a: raise ValueError('invalid contingency table')
    if abcd > MAXN: raise OverflowError('the grand total of contingency table is too large')
    a_min = max(0, ab+ac-abcd)
    a_max = min(ab, ac)
    if a_min == a_max: return 0.
    p0 = lgamma(ab+1) + lgamma(ac+1) + lgamma(abcd-ac+1) + lgamma(abcd-ab+1) - lgamma(abcd+1)
    pa = lgamma(a+1) + lgamma(ab-a+1) + lgamma(ac-a+1) + lgamma(abcd-ab-ac+a+1)
    st = 1.
    if ab * ac < a * abcd:
        for i in range(min(a-1, int(round(ab * ac / abcd))), a_min-1, -1):
            pi = lgamma(i+1) + lgamma(ab-i+1) + lgamma(ac-i+1) + lgamma(abcd-ab-ac+i+1)
            if pi < pa: continue
            st_new = st + exp(pa - pi)
            if st_new == st: break
            st = st_new
        for i in range(a+1, a_max+1):
            pi = lgamma(i+1) + lgamma(ab-i+1) + lgamma(ac-i+1) + lgamma(abcd-ab-ac+i+1)
            st_new = st + exp(pa - pi)
            if st_new == st: break
            st = st_new
    else:
        for i in range(a-1, a_min-1, -1):
            pi = lgamma(i+1) + lgamma(ab-i+1) + lgamma(ac-i+1) + lgamma(abcd-ab-ac+i+1)
            st_new = st + exp(pa - pi)
            if st_new == st: break
            st = st_new
        for i in range(max(a+1, int(round(ab * ac / abcd))), a_max+1):
            pi = lgamma(i+1) + lgamma(ab-i+1) + lgamma(ac-i+1) + lgamma(abcd-ab-ac+i+1)
            if pi < pa: continue
            st_new = st + exp(pa - pi)
            if st_new == st: break
            st = st_new
    return max(0, pa - p0 - log(st))

def mlog10Test1t(a, b, c, d):
    return mlnTest2t(a, a+b, a+c, a+b+c+d)/LN10

def mlog10Test2t(a, ab, ac, abcd):
    return mlnTest2t(a, ab, ac, abcd)/LN10



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="collocation", # Replace with your own username
    version="0.0.7",
    author="Yongfu Liao",
    author_email="liao961120@gmail.com",
    description="Collocation extraction from segmented texts",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/liao961120/collocation",
    # package_dir = {'': 'src'},
    packages=['collocation'],
    # package_data={
    #     "": ["../data/cilin_tree.json"],
    # },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)